"""
Factory functions for technical analysis indicators.

Indicator factory names are all upper case.

Indicators offer a composable interface where a calculation routine
is bound together with its calculation parameters.

An indicator object is a callable that can be applied to prices or series data.

Indicators can be chained with the `@` operator as in `ROC(1) @ SMA(20)`.

The `@` operator can also be used to apply an indicator to its parameter.

So for example `SMA(50) @ prices` can be used to compute the 50 period simple moving average on `prices`,
instead of the more verbose `SMA(50)(prices)`.
"""

# Do not edit! This file was generated by make-indicators.ipynb

from . import core
from .core import wrap_indicator
from .model import FuncIndicator

nan = float('nan')




@wrap_indicator(core.calc_price)
def PRICE(item: str = None):
    params = dict(item=item)
    return FuncIndicator('PRICE', core.calc_price, params=params)


@wrap_indicator(core.calc_avgprice)
def AVGPRICE():
    params = dict()
    return FuncIndicator('AVGPRICE', core.calc_avgprice, params=params)


@wrap_indicator(core.calc_typprice)
def TYPPRICE():
    params = dict()
    return FuncIndicator('TYPPRICE', core.calc_typprice, params=params)


@wrap_indicator(core.calc_wclprice)
def WCLPRICE():
    params = dict()
    return FuncIndicator('WCLPRICE', core.calc_wclprice, params=params)


@wrap_indicator(core.calc_midprice)
def MIDPRICE():
    params = dict()
    return FuncIndicator('MIDPRICE', core.calc_midprice, params=params)


@wrap_indicator(core.calc_crossover)
def CROSSOVER(level: float = 0.0, *, item: str = None):
    params = dict(level=level, item=item)
    return FuncIndicator('CROSSOVER', core.calc_crossover, params=params)


@wrap_indicator(core.calc_crossunder)
def CROSSUNDER(level: float = 0.0, *, item: str = None):
    params = dict(level=level, item=item)
    return FuncIndicator('CROSSUNDER', core.calc_crossunder, params=params)


@wrap_indicator(core.calc_flag)
def FLAG(*, item: str = None):
    params = dict(item=item)
    return FuncIndicator('FLAG', core.calc_flag, params=params)


@wrap_indicator(core.calc_updown)
def UPDOWN(up_level: float = 0.0, down_level: float = 0.0, *, item: str = None):
    params = dict(up_level=up_level, down_level=down_level, item=item)
    return FuncIndicator('UPDOWN', core.calc_updown, params=params)


@wrap_indicator(core.calc_sign)
def SIGN(na_value: float = nan, *, item: str = None):
    params = dict(na_value=na_value, item=item)
    return FuncIndicator('SIGN', core.calc_sign, params=params)


@wrap_indicator(core.calc_step)
def STEP(threshold: 'float' = 1.0, *, item: str = None):
    params = dict(threshold=threshold, item=item)
    return FuncIndicator('STEP', core.calc_step, params=params)


@wrap_indicator(core.calc_clag)
def CLAG(period: int = 1, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('CLAG', core.calc_clag, params=params)


@wrap_indicator(core.calc_abs)
def ABS(*, item: str = None):
    params = dict(item=item)
    return FuncIndicator('ABS', core.calc_abs, params=params)


@wrap_indicator(core.calc_log)
def LOG(*, item: str = None):
    params = dict(item=item)
    return FuncIndicator('LOG', core.calc_log, params=params)


@wrap_indicator(core.calc_exp)
def EXP(*, item: str = None):
    params = dict(item=item)
    return FuncIndicator('EXP', core.calc_exp, params=params)


@wrap_indicator(core.calc_shift)
def SHIFT(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('SHIFT', core.calc_shift, params=params)


@wrap_indicator(core.calc_diff)
def DIFF(period: int = 1, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('DIFF', core.calc_diff, params=params)


@wrap_indicator(core.calc_lag)
def LAG(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('LAG', core.calc_lag, params=params)


@wrap_indicator(core.calc_min)
def MIN(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('MIN', core.calc_min, params=params)


@wrap_indicator(core.calc_max)
def MAX(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('MAX', core.calc_max, params=params)


@wrap_indicator(core.calc_sum)
def SUM(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('SUM', core.calc_sum, params=params)


@wrap_indicator(core.calc_roc)
def ROC(period: int = 1, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('ROC', core.calc_roc, params=params)


@wrap_indicator(core.calc_lroc)
def LROC(period: int = 1, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('LROC', core.calc_lroc, params=params)


@wrap_indicator(core.calc_mad)
def MAD(period: int = 14, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('MAD', core.calc_mad, params=params)


@wrap_indicator(core.calc_stdev)
def STDEV(period: int = 20, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('STDEV', core.calc_stdev, params=params)


@wrap_indicator(core.calc_mav)
def MAV(period: int = 20, *, ma_type: str = 'SMA', item: str = None):
    params = dict(period=period, ma_type=ma_type, item=item)
    return FuncIndicator('MAV', core.calc_mav, params=params)


@wrap_indicator(core.calc_sma)
def SMA(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('SMA', core.calc_sma, params=params)


@wrap_indicator(core.calc_ema)
def EMA(period: int, *, adjust: bool = False, item: str = None):
    params = dict(period=period, adjust=adjust, item=item)
    return FuncIndicator('EMA', core.calc_ema, params=params)


@wrap_indicator(core.calc_rma)
def RMA(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('RMA', core.calc_rma, params=params)


@wrap_indicator(core.calc_wma)
def WMA(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('WMA', core.calc_wma, params=params)


@wrap_indicator(core.calc_hma)
def HMA(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('HMA', core.calc_hma, params=params)


@wrap_indicator(core.calc_dema)
def DEMA(period: int, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('DEMA', core.calc_dema, params=params)


@wrap_indicator(core.calc_tema)
def TEMA(period: int = 20, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('TEMA', core.calc_tema, params=params)


@wrap_indicator(core.calc_alma)
def ALMA(period: int = 9, offset: float = 0.85, sigma: float = 6.0, *, item: str = None):
    params = dict(period=period, offset=offset, sigma=sigma, item=item)
    return FuncIndicator('ALMA', core.calc_alma, params=params)


@wrap_indicator(core.calc_rsi)
def RSI(period: int = 14, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('RSI', core.calc_rsi, params=params)


@wrap_indicator(core.calc_dmi)
def DMI(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('DMI', core.calc_dmi, params=params)


@wrap_indicator(core.calc_adx)
def ADX(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('ADX', core.calc_adx, params=params)


@wrap_indicator(core.calc_pdi)
def PDI(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('PDI', core.calc_pdi, params=params)


@wrap_indicator(core.calc_mdi)
def MDI(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('MDI', core.calc_mdi, params=params)


@wrap_indicator(core.calc_trange)
def TRANGE(*, log_prices: bool = False, percent: bool = False):
    params = dict(log_prices=log_prices, percent=percent)
    return FuncIndicator('TRANGE', core.calc_trange, params=params)


@wrap_indicator(core.calc_atr)
def ATR(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('ATR', core.calc_atr, params=params)


@wrap_indicator(core.calc_natr)
def NATR(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('NATR', core.calc_natr, params=params)


@wrap_indicator(core.calc_sar)
def SAR(afs: float = 0.02, maxaf: float = 0.2):
    params = dict(afs=afs, maxaf=maxaf)
    return FuncIndicator('SAR', core.calc_sar, params=params)


@wrap_indicator(core.calc_cci)
def CCI(period: int = 20):
    params = dict(period=period)
    return FuncIndicator('CCI', core.calc_cci, params=params)


@wrap_indicator(core.calc_cmf)
def CMF(period: int = 20):
    params = dict(period=period)
    return FuncIndicator('CMF', core.calc_cmf, params=params)


@wrap_indicator(core.calc_mfi)
def MFI(period: int = 14):
    params = dict(period=period)
    return FuncIndicator('MFI', core.calc_mfi, params=params)


@wrap_indicator(core.calc_bop)
def BOP(period: int = 20):
    params = dict(period=period)
    return FuncIndicator('BOP', core.calc_bop, params=params)


@wrap_indicator(core.calc_bbands)
def BBANDS(period: int = 20, nbdev: float = 2.0):
    params = dict(period=period, nbdev=nbdev)
    return FuncIndicator('BBANDS', core.calc_bbands, params=params)


@wrap_indicator(core.calc_keltner)
def KELTNER(period: int = 20, nbatr: float = 2.0):
    params = dict(period=period, nbatr=nbatr)
    return FuncIndicator('KELTNER', core.calc_keltner, params=params)


@wrap_indicator(core.calc_ker)
def KER(period: int = 10, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('KER', core.calc_ker, params=params)


@wrap_indicator(core.calc_kama)
def KAMA(period: int = 10, fastn: int = 2, slown: int = 30, *, item: str = None):
    params = dict(period=period, fastn=fastn, slown=slown, item=item)
    return FuncIndicator('KAMA', core.calc_kama, params=params)


@wrap_indicator(core.calc_macd)
def MACD(n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None):
    params = dict(n1=n1, n2=n2, n3=n3, item=item)
    return FuncIndicator('MACD', core.calc_macd, params=params)


@wrap_indicator(core.calc_ppo)
def PPO(n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None):
    params = dict(n1=n1, n2=n2, n3=n3, item=item)
    return FuncIndicator('PPO', core.calc_ppo, params=params)


@wrap_indicator(core.calc_slope)
def SLOPE(period: int = 20, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('SLOPE', core.calc_slope, params=params)


@wrap_indicator(core.calc_rvalue)
def RVALUE(period: int = 20, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('RVALUE', core.calc_rvalue, params=params)


@wrap_indicator(core.calc_tsf)
def TSF(period: int = 20, offset: int = 0, *, item: str = None):
    params = dict(period=period, offset=offset, item=item)
    return FuncIndicator('TSF', core.calc_tsf, params=params)


@wrap_indicator(core.calc_curve)
def CURVE(period: int = 20, *, item: str = None):
    params = dict(period=period, item=item)
    return FuncIndicator('CURVE', core.calc_curve, params=params)


@wrap_indicator(core.calc_qsf)
def QSF(period: int = 20, offset: int = 0, *, item: str = None):
    params = dict(period=period, offset=offset, item=item)
    return FuncIndicator('QSF', core.calc_qsf, params=params)


@wrap_indicator(core.calc_stoch)
def STOCH(period: int = 14, fastn: int = 3, slown: int = 3):
    params = dict(period=period, fastn=fastn, slown=slown)
    return FuncIndicator('STOCH', core.calc_stoch, params=params)


@wrap_indicator(core.calc_streak)
def STREAK(*, item: str = None):
    params = dict(item=item)
    return FuncIndicator('STREAK', core.calc_streak, params=params)


@wrap_indicator(core.calc_eval)
def EVAL(expr: str, *, as_flag: bool = False):
    params = dict(expr=expr, as_flag=as_flag)
    return FuncIndicator('EVAL', core.calc_eval, params=params)


__all__ = [
    'PRICE', 'AVGPRICE', 'TYPPRICE', 'WCLPRICE', 'MIDPRICE', 'CROSSOVER',
    'CROSSUNDER', 'FLAG', 'UPDOWN', 'SIGN', 'STEP', 'CLAG', 'ABS', 'LOG',
    'EXP', 'SHIFT', 'DIFF', 'LAG', 'MIN', 'MAX', 'SUM', 'ROC', 'LROC',
    'MAD', 'STDEV', 'MAV', 'SMA', 'EMA', 'RMA', 'WMA', 'HMA', 'DEMA',
    'TEMA', 'ALMA', 'RSI', 'DMI', 'ADX', 'PDI', 'MDI', 'TRANGE', 'ATR',
    'NATR', 'SAR', 'CCI', 'CMF', 'MFI', 'BOP', 'BBANDS', 'KELTNER', 'KER',
    'KAMA', 'MACD', 'PPO', 'SLOPE', 'RVALUE', 'TSF', 'CURVE', 'QSF',
    'STOCH', 'STREAK', 'EVAL'
]



import inspect
import warnings


def annotate_parameter(param):
    if param.annotation in ("int", "long"):
        return param.replace(annotation=int)

    if param.annotation in ("bint", "bool"):
        return param.replace(annotation=bool)

    if param.annotation == "str":
        return param.replace(annotation=str)

    if param.annotation is inspect._empty:
        if type(param.default) in (int, float, bool):
            return param.replace(annotation=type(param.default))

        if param.name in ("expr", "ma_type"):
            return param.replace(annotation=str)

        if param.name == "period":
            return param.replace(annotation=int)

        if param.name not in ("series", "prices"):
            warnings.warn(f"Missing annotation for {param.name}", stacklevel=2)

    return param



"""
Calculation functions for technical analysis indicators.

The function names are all lower case and may conflict with standard functions,
so the best way to use this module is to alias it to a short name
like `ta` and access all functions as attributes.

The first parameter `series` or `prices` indicates whether the function
accepts a single series or a prices dataframe.

Functions that accept a series usually have an optional parameter `item`
to specify which column to use if the input is a dataframe.

All functions wrap their output to match the type of their input.

In particular the result of a function applied to a pandas series or dataframes
will have the same index as the input.
"""

# Do not edit! This file was generated

from . import core
from .core import get_series, wrap_function, wrap_result


nan = float('nan')

def __getattr__(name):
    if name.isupper() and name.lower() in globals():
        return globals()[name.lower()]
    raise AttributeError(f"Module {__name__} has no attribute {name!r}")

__all__ = ()


@wrap_function(core.calc_price)
def price(prices, item: str = None):
    kwargs = dict()
    result = core.calc_price(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_avgprice)
def avgprice(prices):
    kwargs = dict()
    result = core.calc_avgprice(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_typprice)
def typprice(prices):
    kwargs = dict()
    result = core.calc_typprice(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_wclprice)
def wclprice(prices):
    kwargs = dict()
    result = core.calc_wclprice(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_midprice)
def midprice(prices):
    kwargs = dict()
    result = core.calc_midprice(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_crossover)
def crossover(series, level: float = 0.0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(level=level)
    result = core.calc_crossover(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_crossunder)
def crossunder(series, level: float = 0.0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(level=level)
    result = core.calc_crossunder(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_flag)
def flag(series, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict()
    result = core.calc_flag(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_updown)
def updown(series, up_level: float = 0.0, down_level: float = 0.0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(up_level=up_level, down_level=down_level)
    result = core.calc_updown(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_sign)
def sign(series, na_value: float = nan, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(na_value=na_value)
    result = core.calc_sign(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_step)
def step(series, threshold: 'float' = 1.0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(threshold=threshold)
    result = core.calc_step(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_clag)
def clag(series, period: int = 1, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_clag(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_abs)
def abs(series, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict()
    result = core.calc_abs(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_log)
def log(series, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict()
    result = core.calc_log(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_exp)
def exp(series, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict()
    result = core.calc_exp(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_shift)
def shift(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_shift(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_diff)
def diff(series, period: int = 1, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_diff(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_lag)
def lag(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_lag(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_min)
def min(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_min(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_max)
def max(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_max(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_sum)
def sum(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_sum(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_roc)
def roc(series, period: int = 1, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_roc(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_lroc)
def lroc(series, period: int = 1, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_lroc(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_mad)
def mad(series, period: int = 14, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_mad(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_stdev)
def stdev(series, period: int = 20, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_stdev(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_mav)
def mav(series, period: int = 20, *, ma_type: str = 'SMA', item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period, ma_type=ma_type)
    result = core.calc_mav(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_sma)
def sma(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_sma(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_ema)
def ema(series, period: int, *, adjust: bool = False, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period, adjust=adjust)
    result = core.calc_ema(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_rma)
def rma(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_rma(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_wma)
def wma(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_wma(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_hma)
def hma(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_hma(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_dema)
def dema(series, period: int, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_dema(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_tema)
def tema(series, period: int = 20, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_tema(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_alma)
def alma(series, period: int = 9, offset: float = 0.85, sigma: float = 6.0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period, offset=offset, sigma=sigma)
    result = core.calc_alma(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_rsi)
def rsi(series, period: int = 14, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_rsi(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_dmi)
def dmi(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_dmi(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_adx)
def adx(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_adx(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_pdi)
def pdi(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_pdi(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_mdi)
def mdi(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_mdi(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_trange)
def trange(prices, *, log_prices: bool = False, percent: bool = False):
    kwargs = dict(log_prices=log_prices, percent=percent)
    result = core.calc_trange(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_atr)
def atr(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_atr(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_natr)
def natr(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_natr(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_sar)
def sar(prices, afs: float = 0.02, maxaf: float = 0.2):
    kwargs = dict(afs=afs, maxaf=maxaf)
    result = core.calc_sar(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_cci)
def cci(prices, period: int = 20):
    kwargs = dict(period=period)
    result = core.calc_cci(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_cmf)
def cmf(prices, period: int = 20):
    kwargs = dict(period=period)
    result = core.calc_cmf(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_mfi)
def mfi(prices, period: int = 14):
    kwargs = dict(period=period)
    result = core.calc_mfi(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_bop)
def bop(prices, period: int = 20):
    kwargs = dict(period=period)
    result = core.calc_bop(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_bbands)
def bbands(prices, period: int = 20, nbdev: float = 2.0):
    kwargs = dict(period=period, nbdev=nbdev)
    result = core.calc_bbands(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_keltner)
def keltner(prices, period: int = 20, nbatr: float = 2.0):
    kwargs = dict(period=period, nbatr=nbatr)
    result = core.calc_keltner(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_ker)
def ker(series, period: int = 10, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_ker(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_kama)
def kama(series, period: int = 10, fastn: int = 2, slown: int = 30, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period, fastn=fastn, slown=slown)
    result = core.calc_kama(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_macd)
def macd(series, n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(n1=n1, n2=n2, n3=n3)
    result = core.calc_macd(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_ppo)
def ppo(series, n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(n1=n1, n2=n2, n3=n3)
    result = core.calc_ppo(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_slope)
def slope(series, period: int = 20, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_slope(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_rvalue)
def rvalue(series, period: int = 20, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_rvalue(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_tsf)
def tsf(series, period: int = 20, offset: int = 0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period, offset=offset)
    result = core.calc_tsf(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_curve)
def curve(series, period: int = 20, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period)
    result = core.calc_curve(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_qsf)
def qsf(series, period: int = 20, offset: int = 0, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict(period=period, offset=offset)
    result = core.calc_qsf(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_stoch)
def stoch(prices, period: int = 14, fastn: int = 3, slown: int = 3):
    kwargs = dict(period=period, fastn=fastn, slown=slown)
    result = core.calc_stoch(prices, **kwargs)
    return wrap_result(result, prices)


@wrap_function(core.calc_streak)
def streak(series, *, item: str = None):
    series = get_series(series, item=item)
    kwargs = dict()
    result = core.calc_streak(series, **kwargs)
    return wrap_result(result, series)


@wrap_function(core.calc_eval)
def eval(prices, expr: str, *, as_flag: bool = False):
    kwargs = dict(expr=expr, as_flag=as_flag)
    result = core.calc_eval(prices, **kwargs)
    return wrap_result(result, prices)





"""Model classes"""

import copy
import inspect

from typing import Callable
from types import MappingProxyType
from functools import cached_property

from abc import ABCMeta, abstractmethod

from .core import get_series, wrap_result
from .utils import format_partial, lazy_repr


class Indicator(metaclass=ABCMeta):
    """Abstact Base class for Indicators"""

    __repr__ = lazy_repr

    @abstractmethod
    def __call__(self, data): ...

    def __matmul__(self, other):
        if callable(other):
            return ComposedIndicator(self, other)

        return self(other)

    def get_series(self, data):
        """Series data accessor"""
        item = getattr(self, "item", None)
        return get_series(data, item=item)

    def alias(self, name: str):
        return AliasedIndicator(self, name)


class AliasedIndicator(Indicator):
    """Aliased Indicator"""

    def __init__(self, indicator, name):
        self.indicator = indicator
        self.name = name
    
    def __repr__(self):
        return f"{self.indicator!r}.alias({self.name!r})"

    def __call__(self, data):
        result = self.indicator(data)
        rtype = type(result).__name__

        if rtype == "Series" and hasattr(result, "rename"):
            return result.rename(self.name)

        raise ValueError(f"Cannot rename {rtype} result")



class FuncIndicator(Indicator):
    """Function based Indicator"""

    output_name: str = None

    @staticmethod
    def indicator_name(func):
        name = func.__name__
        name = name.removeprefix("calc_")
        name = name.upper()
        return name

    def __init__(self, name: str, func: Callable, params: dict):
        self.name = name
        self.func = func
        self.item = params.pop("item", None)
        self.params = MappingProxyType(params)

        metadata = getattr(func, "metadata", None)
        if metadata:
            self.metadata = MappingProxyType(metadata)

    @cached_property
    def input_type(self):
        signature = inspect.signature(self.func)
        return next(iter(signature.parameters), None)

    def alias_legacy(self, name):
        if hasattr(self, "output_names"):
            raise ValueError("Cannot alias a multi-output indicator")
        target = copy.copy(self)
        target.output_name = name
        return target


    def __repr__(self):
        return format_partial(self.func, self.params, name=self.name)

    def __call__(self, prices):
        output_name = getattr(self, "output_name", None)

        if self.input_type == "series":
            series = get_series(prices, self.item)
            result = self.func(series, **self.params)
        else:
            result = self.func(prices, **self.params)

        return wrap_result(result, prices, name=output_name)


class ComposedIndicator(Indicator):
    """Composition of Indicators"""

    def __init__(self, *chain):
        items = []
        for item in chain:
            if isinstance(item, ComposedIndicator):
                items.extend(item.chain)
            else:
                items.append(item)
        self.chain = tuple(items)

    def __repr__(self):
        return " @ ".join(repr(fn) for fn in self.chain)

    def __call__(self, data):
        for fn in reversed(self.chain):
            data = fn(data)
        return data



"""Mintalib Utils"""

from inspect import Signature, Parameter



def format_partial(func, data, *, name: str = None):
    """format a partial function call"""

    if name is None:
        name = func.__name__

    signature = Signature.from_callable(func)
    positional = True
    arguments = []

    for k, v in data.items():
        p = signature.parameters.get(k)

        if not p or p.kind not in (
            Parameter.POSITIONAL_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
        ):
            positional = False

        if positional:
            arguments.append(f"{v!r}")
        elif not p or v != p.default:
            arguments.append(f"{k!s}={v!r}")

    arguments = ", ".join(arguments)
    return "%s(%s)" % (name, arguments)


def lazy_repr(obj):
    """minimal __repr__ based on __init__ signature"""

    cname = obj.__class__.__qualname__

    return format_partial(obj.__init__, obj.__dict__, name=cname)




import sys
from pathlib import Path
from setuptools import setup, find_packages, Extension

srcdir = "src"
package_dir = {"": srcdir}
packages = find_packages(where=srcdir)
extra_compile_args = []

# Extra compilation flags to suppress cython related warnings on MacOS ...
if sys.platform == "darwin":
    extra_compile_args = ["-Wno-unreachable-code", "-Wno-deprecated-declarations"]

def make_extension(path):
    name = path.relative_to(srcdir).with_suffix("").as_posix().replace("/", ".")
    return Extension(
        name=name, sources=[str(path)], extra_compile_args=extra_compile_args
    )

# ext_modules = [make_extension(f) for f in Path(srcdir).rglob("*.pyx")]
ext_modules = [make_extension(f) for f in Path(srcdir).rglob("*.c")]

setup(
    packages=packages,
    package_dir=package_dir,
    ext_modules=ext_modules,
)
